"""Module containing Models, the main classes of this package which enables the tracing and interleaving functionality of nnsight.

Models allow users to load and wrap torch modules. 

Here we load 'gpt2' from huggingface using its repo id into a LanguageModel:

>>> from nnsight import LanguageModel

>>> model = LanguageModel('gpt2', device_map='cuda:0')

In this case, using a LanguageModel means the underlying model is a transformers.AutoModelForCausalLM, and unused arguments by LanguageModel are passed downstream to AutoModelForCausalLM.
``device_map='cuda:0`` ultimately leverages the accelerate package to use the first gpu when loading the local_model.

The wrapping of the underlying model is further encapsulated by both printing out it's module structure when printing the wrapped model, as well as accessing attributes on the meta_model by accessing them on the wrapped module.

Printing out the wrapped module demonstrates this like:

>>> print(model)

GPT2LMHeadModel(
  (transformer): GPT2Model(
    (wte): Embedding(50257, 768)
    (wpe): Embedding(1024, 768)
    (drop): Dropout(p=0.1, inplace=False)
    (h): ModuleList(
      (0-11): 12 x GPT2Block(
        (ln_1): LayerNorm((768,), eps=1e-05, elementwise_affine=True)
        (attn): GPT2Attention(
          (c_attn): Conv1D()
          (c_proj): Conv1D()
          (attn_dropout): Dropout(p=0.1, inplace=False)
          (resid_dropout): Dropout(p=0.1, inplace=False)
        )
        (ln_2): LayerNorm((768,), eps=1e-05, elementwise_affine=True)
        (mlp): GPT2MLP(
          (c_fc): Conv1D()
          (c_proj): Conv1D()
          (act): NewGELUActivation()
          (dropout): Dropout(p=0.1, inplace=False)
        )
      )
    )
    (ln_f): LayerNorm((768,), eps=1e-05, elementwise_affine=True)
  )
  (lm_head): Linear(in_features=768, out_features=50257, bias=False)
)

The main two methods to interact and run the model are ``.generate(...)`` and ``.forward(...)``

Both return context manager objects which when entered, track operations you perform on the inputs and outputs of modules.

The generate context is meant for multi-iteration, and multi-input inference runs.
Arguments passed to generate determine the generation behavior, in this case to generate three tokens.
Within a generation context, invoker sub-contexts are entered using ``generator.invoke``. This is where input (or batch of inputs) to the model is accepted, and batched with other invocations.
It's in these contexts where operations on inputs and outputs of modules are tracked and prepared for execution.

In this example, we run two prompts on the language model in order to generate three tokens.
We also perform a ``.save()`` operation on the output of the lm_head module, the logit outputs, in order to save these activations and access them after generation is over:

>>> model.generate(max_new_tokens=3) as generator:
>>>     generator.invoke("The Eiffel Tower is in the city of") as invoker:
>>>         logits1 = model.lm_head.output.save()
>>>     generator.invoke("The Empire State Building is in the city of") as invoker:
>>>         logits2 = model.lm_head.output.save()
>>> print(logits1.value)
>>> print(logits2.value)
>>> print(runner.output)

The forward, runner context is meant for direct input to the underlying model (or module) in either inference or training mode where gradients are collected.

>>> model.forward("The Eiffel Tower is in the city of", inference=True) as runner:
>>>     logits = model.lm_head.output.save()
>>> print(logits.value)
>>> print(runner.output)
"""



from __future__ import annotations

from abc import abstractmethod
from contextlib import AbstractContextManager
from typing import TYPE_CHECKING, List
from ..intervention import InterventionProxy

from ..fx.Graph import Graph

if TYPE_CHECKING:
    from ..models.AbstractModel import AbstractModel


class Tracer(AbstractContextManager):
    def __init__(
        self,
        model: "AbstractModel",
        *args,
        **kwargs,
    ) -> None:
        self.model = model

        self.args = args
        self.kwargs = kwargs

        self.graph = Graph(self.model.meta_model, proxy_class=InterventionProxy)

        self.batch_size: int = 0
        self.input_ids: List[List[int]] = []
        self.output = None

        self.generation_idx:int = 0

        # Modules need to know about the current Tracer to create the correct proxies.
        for name, module in self.model.named_modules():
            module.tracer = self

    @abstractmethod
    def __enter__(self) -> Tracer:
        raise NotImplementedError()

    @abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        raise NotImplementedError()



from __future__ import annotations

from typing import Any, Dict, List, Union

import torch

from . import util
from .contexts.Tracer import Tracer
from .fx.Graph import Graph
from .fx.Node import Node
from .intervention import InterventionProxy

"""The Module object acts as the entrypoint into the interwoven intervention graph for manipulating the inputs and outputs of Modules as a model is running.
Specifically, the `.output` and `.input` attributes are the root nodes of the intervention graph, and all operation performed on these Proxy objects are the downstream nodes that populate the graph.
Requires a Tracewr object to add the correct proxies to the intervention graph. 

Examples:

    Below example shows accessing the input and output of a gpt2 module, saving them during execution, and printing the resulting values after execution:

    >>> with model.generate() as generator:
    >>>    with generator.invoke('The Eiffel Tower is in the city of') as invoker:
    >>>        hidden_states = model.lm_head.input.save()
    >>>        logits = model.lm_head.output.save()
    >>> print(hidden_states.value)
    >>> print(logits.value)

    Below example shows accessing the output of a gpt2 module and setting the values to zero:

    >>> with model.generate() as generator:
    >>>    with generator.invoke('The Eiffel Tower is in the city of') as invoker:
    >>>        model.transformer.h[0].output[0] = 0

"""


class Module(torch.nn.Module):
    """Class meant to wrap existing torch modules within a model's module tree in order to add nnsight functionality.
    Proxies of it's output and input are accessed by `.output` and `.input` respectively.

    Attributes:
        module_path (str): String representing the attribute path of this module relative the the root model. Seperarated by '.' e.x ('transformer.h.0.mlp'). Set by AbstractModel on intialization of meta model.
        output (nnsight.intervention.InterventionProxy): Proxy object representing the output of this module. Reset on pass through.
        output_shape (torch.Size): Shape of the tensor outputs to this module. Populated by most recent pass through. Can also be a nested list of torch.Size.
        output_type (torch.dtype): Dtype of the tensor outputs to this module. Populated by most recent pass through. Can also be a nested list of torch.dtype.
        output (nnsight.intervention.InterventionProxy): Proxy object representing the output of this module. Reset on pass through.
        input_shape (torch.Size): Shape of the tensor inputs to this module. Populated by most recent pass through. Can also be a nested list of torch.Size.
        input_type (torch.dtype): Dtype of the tensor inputs to this module. Populated by most recent pass through. Can also be a nested list of torch.dtype.
        tracer (nnsight.context.Tracer.Tracer): Object which adds this module's output and input proxies to an intervention graph. Must be set on Module objects manually.
    """

    def __init__(self) -> None:
        self.module_path: str = None
        self.input_shape: torch.Size = None
        self.input_type: torch.dtype = None
        self.output_shape: torch.Size = None
        self.output_type: torch.dtype = None

        self._output: InterventionProxy = None
        self._input: InterventionProxy = None
        self._graph: Graph = None

        self.tracer: Tracer = None

    def __call__(
        self, *args: List[Any], **kwds: Dict[str, Any]
    ) -> Union[Any, InterventionProxy]:
        """Override __call__ to check for InterventionProxy arguments.
        If there are any, we should return an InterventionProxy denoting we want to call the given module with arguments.

        Args:
            args (List[Any]): Positional arguments.
            kwargs (Dict[str,Any]): Keyword arguments

        Returns:
            Union[Any,InterventionProxy]: Either the output of the module if not tracing or a Proxy if tracing.
        """
        proxy = any(
            isinstance(x, InterventionProxy) for x in list(args) + list(kwds.values())
        )

        if proxy:
            module_proxy = getattr(self.tracer.graph.module_proxy, self.module_path)

            return module_proxy.forward(*args, **kwds)

        return super().__call__(*args, **kwds)

    @property
    def output(self) -> InterventionProxy:
        """
        Calling denotes the user wishes to get the output of this module and therefore we create a Proxy of that request.
        Only generates a proxy the first time it is references otherwise return the already set one.

        Returns:
            Proxy: Output proxy.
        """
        if self._output is None:
            self._output = self.tracer.graph.add(
                graph=self.tracer.graph,
                value=util.apply(
                    self.output_shape,
                    lambda x: torch.empty(x, device="meta"),
                    torch.Size,
                ),
                target="argument",
                args=[
                    f"{self.module_path}.output.{self.tracer.generation_idx}",
                    self.tracer.batch_size,
                    len(self.tracer.input_ids) - self.tracer.batch_size,
                ],
            )

        return self._output

    @output.setter
    def output(self, value: Union[InterventionProxy, Any]) -> None:
        """
        Calling denotes the user wishes to set the output of this module and therefore we create a Proxy of that request.

        Args:
            value (Union[Proxy, Any]): _description_
        """

        Node.update(
            self.output.node.proxy_value, self.output.node.prepare_proxy_values(value)
        )

        self.output.node.graph.add(
            graph=self.output.node.graph,
            value=self.output.node.proxy_value,
            target=Node.update,
            args=[self.output.node, value],
        )

    @property
    def input(self) -> InterventionProxy:
        """
        Calling denotes the user wishes to get the input of this module and therefore we create a Proxy of that request.
        Only generates a proxy the first time it is references otherwise return the already set one.

        Returns:
            Proxy: Input proxy.
        """
        if self._input is None:
            self._input = self.tracer.graph.add(
                graph=self.tracer.graph,
                value=util.apply(
                    self.input_shape,
                    lambda x: torch.empty(x, device="meta"),
                    torch.Size,
                ),
                target="argument",
                args=[
                    f"{self.module_path}.input.{self.tracer.generation_idx}",
                    self.tracer.batch_size,
                    len(self.tracer.input_ids) - self.tracer.batch_size,
                ],
            )

        return self._input

    @input.setter
    def input(self, value: Union[InterventionProxy, Any]) -> None:
        """
        Calling denotes the user wishes to set the input of this module and therefore we create a Proxy of that request.

        Args:
            value (Union[Proxy, Any]): _description_
        """

        Node.update(
            self.input.node.proxy_value, self.input.node.prepare_proxy_values(value)
        )

        self.input.node.graph.add(
            graph=self.input.node.graph,
            value=self.input.node.proxy_value,
            target=Node.update,
            args=[self.input.node, value],
        )

    @property
    def graph(self) -> Graph:
        if self._graph is None:
            self._graph = Graph.trace(
                self,
                *util.apply(
                    self.input_shape,
                    lambda x: torch.empty(x, device="meta"),
                    torch.Size,
                ),
            )

        return self._graph

    @staticmethod
    def wrap(module: torch.nn.Module) -> Module:
        """Wraps the torch Module with our Module class. Adds a forward hook to the module to populate output and input data. Does not wrap torch.nn.ModuleList.

        Args:
            module (torch.nn.Module): The torch Module to wrap.

        Returns:
            Module: The wrapped Module.
        """

        def hook(module: Module, input: Any, output: Any):
            module._output = None
            module._input = None
            module.output_shape = util.apply(output, lambda x: x.shape, torch.Tensor)
            module.input_shape = util.apply(input, lambda x: x.shape, torch.Tensor)
            module.output_type = util.apply(output, lambda x: x.dtype, torch.Tensor)
            module.input_type = util.apply(input, lambda x: x.dtype, torch.Tensor)

        for name, _module in module.named_children():
            setattr(module, name, Module.wrap(_module))

        if isinstance(module, (Module, torch.nn.ModuleList)):
            return module

        util.wrap(module, Module)

        module.register_forward_hook(hook)

        return module



from __future__ import annotations

import copy
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Union

import accelerate
import torch
from torch.utils.hooks import RemovableHandle

from ..alteration import REPOID_TO_ALTERATION
from ..contexts.Generator import Generator
from ..contexts.Runner import Runner
from ..editing.Editor import Edit, Editor
from ..editing.GraphEdit import GraphEdit
from ..editing.WrapperModuleEdit import WrapperModuleEdit
from ..fx.Graph import Graph
from ..intervention import HookModel, intervene
from ..logger import logger
from ..Module import Module
from ..patching import Patcher


class AbstractModel(ABC):
    """Abstract class to be implmented for pytorch models wishing to gain this package's functionality.

    Attributes:
        repoid_path_clsname (str): Hugging face repo id of model to load, path to checkpoint, or class name of custom model.
        args (List[Any]): Positional arguments used to initialize model.
        kwargs (Dict[str,Any]): Keyword arguments used to initialize model.
        alter (bool): If to check for alterations and apply them to the model.
        dispatched (bool): If the local_model has bee loaded yet.
        custom_model (bool): If the value passed to repoid_path_model was a custom model.
        meta_model (nnsight.Module): Version of the root model where all parameters and tensors are on the 'meta'
            device. All modules are wrapped in nnsight.Module adding interleaving operation functionality.
        tokenizer (_typ_): Tokenizer.
        local_model (torch.nn.Module): Locally loaded and dispatched model. Only loaded and dispatched on first use.
            This is the actual model that is ran with hooks added to it to enter the intervetnion graph.
    """

    def __init__(
        self,
        repoid_path_model: Union[str, torch.nn.Module],
        *args,
        alter: bool = True,
        tokenizer=None,
        **kwargs,
    ) -> None:
        super().__init__()

        self.repoid_path_clsname = repoid_path_model
        self.args = args
        self.kwargs = kwargs
        self.alter = alter
        self.dispatched = False
        self.custom_model = False
        self.meta_model: Module = None
        self.tokenizer = tokenizer
        self.local_model: torch.nn.Module = None
        self.edits: List[Edit] = list()

        if isinstance(repoid_path_model, torch.nn.Module):
            self.repoid_path_clsname = repoid_path_model.__class__.__name__
            self.custom_model = True
            self.dispatched = True
            self.local_model = repoid_path_model

        logger.debug(f"Initializing `{self.repoid_path_clsname}`...")

        # If alter and alteration exist, use alteration patcher context while loading module.
        with self.alteration() if self.alter else Patcher():
            # Use accelerate and .to('meta') to assure tensors are loaded to 'meta' device
            with accelerate.init_empty_weights(include_buffers=True):
                if self.custom_model:
                    self.meta_model: Module = Module.wrap(
                        copy.deepcopy(self.local_model).to("meta")
                    )
                else:
                    self.meta_model: Module = Module.wrap(
                        self._load_meta(self.repoid_path_clsname, *args, **kwargs).to(
                            "meta"
                        )
                    )

        # Wrap all modules in our Module class.
        for name, module in self.meta_model.named_children():
            module = Module.wrap(module)

            setattr(self.meta_model, name, module)

        # Set module_path attribute so Modules know their place.
        for name, module in self.meta_model.named_modules():
            module.module_path = name

        # Run inital dummy string to populate Module shapes, dtypes etc
        self._run_meta("_")

        logger.debug(f"Initialized `{self.repoid_path_clsname}`")

    def __repr__(self) -> str:
        return repr(self.meta_model)

    def __getattr__(self, key: Any) -> Any:
        """Allows access of sub-modules on meta_model directly from AbstractModel object

        Args:
            key (Any): Key.

        Returns:
            Any: Attribute.
        """
        return getattr(self.meta_model, key)

    def __call__(
        self,
        fn: Callable,
        inputs: Any,
        graph: Graph,
        *args,
        edits: List[Edit] = None,
        inference: bool = True,
        **kwargs,
    ) -> Any:
        """Runs some function with some inputs and some graph with the approriate contexts for this model.

        Loads and dispatched local_model if not already done so.

        Args:
            fn (Callable): Function to run.
            inputs (Any): Inputs to give to function.
            graph (Graph): Intervention graph to interleave with model's computation graph.
            inference (bool): If running in inference mode. Defaults to True.

        Returns:
            Any: _description_
        """
        if edits is None:
            edits = self.edits

        # If local_model not yet loaded, do so.
        if not self.dispatched:
            with self.alteration() if self.alter else Patcher():
                self.local_model = self._load_local(
                    self.repoid_path_clsname, *self.args, **self.kwargs
                )

                # By default, all params should be frozen.
                for param in self.local_model.parameters():
                    param.requires_grad = False

        with Editor(self, edits):
            # Send local_model to graph to re-compile

            increment_hook = self._register_increment_hook(
                lambda module, input, output: graph.increment()
            )

            # The intervention graph for running a Model will have the modules that are involved
            # in the graph's argument_node_names.
            modules = set(
                [
                    ".".join(name.split(".")[:-2])
                    for name in graph.argument_node_names.keys()
                ]
            )

            logger.debug(f"Running `{self.repoid_path_clsname}`...")

            graph.compile(self.local_model)

            self.local_model.eval() if inference else self.local_model.train()

            with torch.inference_mode(mode=inference):
                with HookModel(
                    self.local_model,
                    list(modules),
                    input_hook=lambda activations, module_path: intervene(
                        activations, module_path, graph, "input"
                    ),
                    output_hook=lambda activations, module_path: intervene(
                        activations, module_path, graph, "output"
                    ),
                ):
                    output = fn(inputs, *args, **kwargs)

            increment_hook.remove()

            logger.debug(f"Completed `{self.repoid_path_clsname}`")

        return output

    def generate(self, *args, **kwargs) -> Generator:
        """Returns a Generator context for this model.

        Generator contexts are used to trace and interleave operations on the model's computation graph over some iterative generation process.

        Arguments passed to ``.generate()`` are passed downstream to the model specific _generation method.

        Generator's are used in tandem with their Invoker contexts to enter inputs for operation tracing and execution.

        The output of the generation is ultimately saved in the generator's ``.output`` attribute.

        Returns:
            Generator: Generator.

        Examples:

            A simple entering of a generation context on a language model, and running a prompt with no interventions:

            >>> with model.generate(max_new_tokens=1) as generator:
            >>>     with generator.invoke('The Eiffel Tower is in the city of') as invoker:
            >>>         pass
            >>> print(generator.output)

            Keyword arguments like 'max_new_tokens' are model specific and in this case limits the amount of tokens to predict to one.

            See the Generator docs for more information.

        """
        return Generator(self, *args, **kwargs)

    def forward(self, inputs, *args, **kwargs) -> Runner:
        """Returns a Runner context for this model.

        Runner contexts are used to trace and interleave operations on the model's computation graph over a single input directly to the underlying model.

        Arguments passed to ``.forward()`` are passed downstream to the model specific _run_local method.

        The output of the runner is ultimately saved in the runner's ``.output`` attribute.

        Returns:
            Runner: Runner.

        Example:

            A simple entering of a runner context on a language model, and running a prompt with no interventions:

            >>> with model.forward('The Eiffel Tower is in the city of') as invoker:
            >>>         pass
            >>> print(invoker.output)
        """
        return Runner(inputs, self, *args, **kwargs)

    def alteration(self) -> Patcher:
        return REPOID_TO_ALTERATION.get(self.repoid_path_clsname, Patcher())

    def modulize(self, module: Module, node_name: str, module_name: str) -> None:
        """_summary_

        Args:
            module (Module): _description_
            node_name (str): _description_
            module_name (str): _description_
        """

        # Create a WrapperModuleEdit which just adds a WrapperModule to an existing module at the given moduel_name.
        wme = WrapperModuleEdit(module.module_path, module_name)
        # Wrap with our Module and update new attributes.
        wme.wrapper: Module = Module.wrap(wme.wrapper)
        wme.wrapper.module_path = f"{module.module_path}.{module_name}"
        wme.wrapper.tracer = module.tracer
        wme.wrapper.output_shape = module.output_shape
        # Carry out the edit on the meta_model.
        wme.edit(self.meta_model)

        # Get/create the execution graph for the module's forward method.
        graph = module.graph

        # Add two proxies/nodes, one to get the new WrapperModule we added and another to call it with the data from the original module.
        # Passing the data through the wrapper module allows hooking of the module's output like usual.
        module_proxy = getattr(graph.module_proxy, module_name)
        module_proxy(graph.nodes[node_name])

        # Create and carry out the edit on the meta_model.
        ge = GraphEdit(module.module_path, module.graph)
        ge.edit(self.meta_model)

        # Append to self.edits so when we call the local model, we temporarily edit the module in the same way as the meta model.
        self.edits.append(wme)
        self.edits.append(ge)

    @abstractmethod
    def _prepare_inputs(self, inputs: Any, **kwargs) -> Any:
        """Abstract method which prepares inputs. To be implemented by inheritors.

        Args:
            inputs (Any): Inputs.

        Returns:
            Any: Prepared inputs.
        """
        raise NotImplementedError()

    @abstractmethod
    def _load_meta(self, repoid_or_path: str, *args, **kwargs) -> torch.nn.Module:
        """
        Abstract method to initialize meta_model and tokenizer. To be implemented by inheritors.

        Args:
            repoid_or_path (str): Huggigface repo id or path to checkpoint.

        Returns:
            torch.nn.Module: Meta version of model
        """
        raise NotImplementedError()

    @abstractmethod
    def _load_local(self, repoid_or_path, *args, **kwargs) -> torch.nn.Module:
        """
        Abstract method to initialize and dispatch the local_model. To be implemented by inheritors.

        Args:
            repoid_or_path (str): Huggigface repo id or path to checkpoint.

        Returns:
            torch.nn.Module: Local version of model
        """
        raise NotImplementedError()

    @abstractmethod
    def _run_meta(self, inputs, *args, **kwargs) -> Any:
        """
        Abstract method to directly call the meta_model. To be implemented by inheritors.

        Used for tracing operations and their input/output shapes/dtypes.

        Args:
            inputs (Any): Inputs.

        Returns:
            Any: Output.
        """
        raise NotImplementedError()

    @abstractmethod
    def _run_local(self, inputs, *args, **kwargs) -> Any:
        """
        Abstract method to directly call the local_model. To be implemented by inheritors.

        Args:
            inputs (Any): Inputs.

        Returns:
            Any: Output.
        """
        raise NotImplementedError()

    @abstractmethod
    def _generation(self, inputs, *args, **kwargs) -> Any:
        """
        Abstract method to do interative generation on the local_model. To be implemented by inheritors.

        Args:
            inputs (Any): Inputs.

        Returns:
            Any: Output.
        """
        raise NotImplementedError()

    @abstractmethod
    def _register_increment_hook(self, hook: Callable) -> RemovableHandle:
        """Abstract method to hook a function on the local_model on the main module that is incremented during generation.

        Args:
            hook (Callable): Increment hook. Probably from the Generator context.

        Returns:
            RemovableHandle: Handle to remove the applied hook after generation is done.
        """
        raise NotImplementedError()



from __future__ import annotations

from .Tracer import Tracer


class Runner(Tracer):
    def __init__(
        self,
        input,
        *args,
        inference:bool=True,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)

        self.input = input
        self.inference = inference

    def __enter__(self) -> Runner:
        token_ids = self.model._run_meta(self.input, *self.args, **self.kwargs)

        # Decode tokenized inputs for user usage.
        self.tokens = [
            [self.model.tokenizer.decode(token) for token in ids] for ids in token_ids
        ]
        self.ids = token_ids

        self.batch_size = len(self.ids)

        self.input_ids.extend(token_ids)

        if len(self.tokens) == 1:
            self.tokens = self.tokens[0]
            self.ids = self.ids[0]

        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.output = self.model(
            self.model._run_local,
            self.input,
            self.graph,
            *self.args,
            inference=self.inference,
            **self.kwargs,
        )
